
#include "XC_Core.h"
#include "XC_Networking.h"
#include "XC_Download.h"
#include "UDCodecs.h"
#include "XC_LZMA.h"


XC_CORE_API UBOOL b440Net = 0;

void FOutBunch_Hack::Init()
{
	b440 = (b440Net != 0);
}

//Editing across all versions
FOutBunchHeader* FOutBunch_Hack::Bunch()
{
	return (FOutBunchHeader*)(((BYTE*)this) + 48 + b440Net*8);
}
//Accessor
FArchiveHeader* FOutBunch_Hack::Archive()
{
	return (FArchiveHeader*)this;
}

//Prevent v436 destructor from killing the app
void FOutBunch_Hack::Exit()
{
	FOutBunchHeader* Header = Bunch();
	if ( Header->Buffer.GetData() )
		Header->Buffer.Empty();
	appMemzero( ((BYTE*)this)+48, 12); 
}


/*=============================================================================
XC_Core extended download protocols
=============================================================================*/

void UXC_Download::StaticConstructor()
{
	EnableLZMA = 1;
	UseCompression = 1;
	
	if ( UNetConnection::StaticClass()->GetPropertiesSize() == 16088 )
		b440Net = 1;
}

void UXC_Download::ReceiveData( BYTE* Data, INT Count )
{
	guard( UXC_Download:ReceiveData);
	// Receiving spooled file data.
	if( Transfered==0 && !RecvFileAr )
	{
		// Open temporary file initially.
		debugf( NAME_DevNet, TEXT("Receiving package '%s'"), Info->Parent->GetName() );
		appCreateTempFilename( *GSys->CachePath, TempFilename );
		FixFilename( TempFilename); //Also initializes linux FM
		GXFileManager->MakeDirectory( *GSys->CachePath, 0 );
		RecvFileAr = GXFileManager->CreateFileWriter( TempFilename );
		if ( Count >= 13 )
		{
			QWORD* LZMASize = (QWORD*)&Data[5];
			if ( Info->FileSize == *LZMASize )
			{
				IsCompressed = 1;
				IsLZMA = 1;
				debugf( NAME_DevNet, TEXT("USES LZMA"));
			}
			INT* UzSignature = (INT*)&Data[0];
			if ( *UzSignature == 1234 || *UzSignature == 5678 )
			{
				IsCompressed = 1;
				debugf( NAME_DevNet, TEXT("USES UZ: Signature %i"), *UzSignature);
			}
		}
	}

	// Receive.
	if( !RecvFileAr )
	{
		// Opening file failed.
		DownloadError( LocalizeError(TEXT("NetOpen"),TEXT("Engine")) );
	}
	else
	{
		if( Count > 0 )
			RecvFileAr->Serialize( Data, Count );
		if( RecvFileAr->IsError() )
		{
			// Write failed.
			DownloadError( *FString::Printf( LocalizeError(TEXT("NetWrite"),TEXT("Engine")), TempFilename ) );
		}
		else
		{
			// Successful.
			Transfered += Count;
			INT RealSize = CompressedSize ? CompressedSize : Info->FileSize;
			FString Msg1 = FString::Printf( (Info->PackageFlags&PKG_ClientOptional)?LocalizeProgress(TEXT("ReceiveOptionalFile"),TEXT("Engine")):LocalizeProgress(TEXT("ReceiveFile"),TEXT("Engine")), Info->Parent->GetName() );
			FString Msg2 = FString::Printf( LocalizeProgress(TEXT("ReceiveSize"),TEXT("Engine")), RealSize/1024, 100.f*Transfered/RealSize );
			Connection->Driver->Notify->NotifyProgress( *Msg1, *Msg2, 4.f );
		}
	}	
	unguard;
}

void UXC_Download::DownloadDone()
{
	guard( UXC_Download::DownloadDone);
	FixFilename( TempFilename);
	if( RecvFileAr )
	{
		guard( DeleteFile );
#if VERBOSE_BUILD
		debugf( NAME_DevNet, TEXT("DownloadDone: deleting file archive") );
#endif
		delete RecvFileAr;
		RecvFileAr = NULL;
		unguard;
	}
	if( SkippedFile )
	{
		guard( Skip );
		debugf( TEXT("Skipped download of '%s'"), Info->Parent->GetName() );
		GXFileManager->Delete( TempFilename );
		TCHAR Msg[256];
		appSprintf( Msg, TEXT("Skipped '%s'"), Info->Parent->GetName() );
		Connection->Driver->Notify->NotifyProgress( TEXT("Success"), Msg, 4.f );
		Connection->Driver->Notify->NotifyReceivedFile( Connection, PackageIndex, TEXT(""), 1 );
		unguard;
	}
	else
	{
		UChannel** Channels = !b440Net ? (UChannel**) &Connection->Channels : (UChannel**) ( (DWORD)Connection + 0x0E74 );
		if ( !Channels[0] || Channels[0]->Closing )
			return;
		FString Dest = GSys->CachePath + PATH_SEPARATOR;
		Dest += Info->Guid.String();
		Dest += GSys->CacheExt;
		//		FString Dest = GSys->CachePath + PATH_SEPARATOR + Info->Guid.String() + GSys->CacheExt;
		if( !Error[0] && Transfered==0 )
			DownloadError( *FString::Printf( LocalizeError(TEXT("NetRefused"),TEXT("Engine")), Info->Parent->GetName() ) );
		if( !Error[0] && IsCompressed )
		{
			TCHAR CFilename[256];
			appStrcpy( CFilename, TempFilename );
			appCreateTempFilename( *GSys->CachePath, TempFilename );
			FArchive* CFileAr = GXFileManager->CreateFileReader( CFilename );
			FArchive* UFileAr = NULL; //Don't open yet
			if ( CFileAr && !IsLZMA )
				UFileAr = GXFileManager->CreateFileWriter( TempFilename );

			if( !CFileAr || (!IsLZMA && !UFileAr) )
				DownloadError( LocalizeError(TEXT("NetOpen"),TEXT("Engine")) );
			else if ( IsLZMA )
			{
				if ( LzmaDecompress( CFileAr, *Dest, Error) )
					debugf( NAME_DevNet, TEXT("LZMA Decompress: %s"), TempFilename);
			}
			else
			{
				INT Signature;
				FString OrigFilename;
				*CFileAr << Signature;
				if( (Signature != 5678) && (Signature != 1234) )
					DownloadError( LocalizeError(TEXT("NetSize"),TEXT("Engine")) );
				else
				{
					*CFileAr << OrigFilename;
					FFCodecFull Codec;
					Codec.AddCodec(new FFCodecRLE);
					Codec.AddCodec(new FFCodecBWT);
					Codec.AddCodec(new FFCodecMTF);
					if ( Signature == 5678 ) //UZ2 Support
						Codec.AddCodec(new FFCodecRLE);
					Codec.AddCodec(new FFCodecHuffman);
					Codec.Decode( *CFileAr, *UFileAr );
				}
			}
			if( CFileAr )
			{
				GXFileManager->Delete( CFilename );
				delete CFileAr;
			}
			if( UFileAr )
				delete UFileAr;
		}
		FixFilename( TempFilename);
		FixFilename( *Dest);
		if( !Error[0] && !IsCompressed && GXFileManager->FileSize(TempFilename)!=Info->FileSize ) //Compression screws up filesize, ignore
			DownloadError( LocalizeError(TEXT("NetSize"),TEXT("Engine")) );
		if( !Error[0] && !IsLZMA && !GXFileManager->Move( *Dest, TempFilename, 1 ) ) //LZMA already performs this step
			DownloadError( LocalizeError(TEXT("NetMove"),TEXT("Engine")) );
		if( Error[0] )
		{
			GXFileManager->Delete( TempFilename );
//HIGOR: Control channel is being closed, do not notify level of file failure (it will restart download using a different method!)
#if VERBOSE_BUILD
			debugf(NAME_DevNet, TEXT("DownloadDone: failed, using different method if possible"));
#endif
			Connection->Driver->Notify->NotifyReceivedFile( Connection, PackageIndex, Error, 0 );
		}
		else
		{
			// Success.
			TCHAR Msg[256];
//			TCHAR IniName[256];
			FString IniName = GSys->CachePath + PATH_SEPARATOR + TEXT("cache.ini");
			FConfigCacheIni CacheIni;
//			appSprintf( IniName, TEXT("%s") PATH_SEPARATOR TEXT("cache.ini"), *GSys->CachePath );
			CacheIni.SetString( TEXT("Cache"), Info->Guid.String(), *(*Info->URL) ? *Info->URL : Info->Parent->GetName(), *IniName );

			appSprintf( Msg, TEXT("Received '%s'"), Info->Parent->GetName() );
			Connection->Driver->Notify->NotifyProgress( TEXT("Success"), Msg, 4.f );
			Connection->Driver->Notify->NotifyReceivedFile( Connection, PackageIndex, Error, 0 );
		}
	}
	unguard;
}
IMPLEMENT_CLASS(UXC_Download)


void UXC_ChannelDownload::StaticConstructor()
{
	DownloadParams = TEXT("Enabled");
	UChannel::ChannelClasses[7] = UXC_FileChannel::StaticClass();
}
void UXC_ChannelDownload::Serialize( FArchive& Ar )
{
	Super::Serialize( Ar );
	Ar << Ch;
}
UBOOL UXC_ChannelDownload::TrySkipFile()
{
	if( Ch && Super::TrySkipFile() )
	{
		FOutBunch_Hack Bunch( Ch, 1 );
		FString Cmd = TEXT("SKIP");
		Bunch << Cmd;
		Bunch.Bunch()->bReliable = 1;
		Ch->SendBunch( &Bunch, 0 );
		Bunch.Exit();
		return 1;
	}
	return 0;
}
void UXC_ChannelDownload::ReceiveFile( UNetConnection* InConnection, INT InPackageIndex, const TCHAR *Params, UBOOL InCompression )
{
	UXC_Download::ReceiveFile( InConnection, InPackageIndex, Params, InCompression );

	// Create channel.
	Ch = (UFileChannel *)Connection->CreateChannel( (EChannelType)7, 1 );

	if( !Ch )
	{
		DownloadError( LocalizeError(TEXT("ChAllocate"),TEXT("Engine")) );
		DownloadDone();
		return;
	}

	// Set channel properties.
	Ch->Download = (UChannelDownload*)this; //THIS IS A HACK!!!!
	Ch->PackageIndex = PackageIndex;

	// Send file request.
	FOutBunch_Hack Bunch( Ch, 0 );
	Bunch.Bunch()->ChType = 7;
	Bunch << Info->Guid;
	Bunch.Bunch()->bReliable = 1;
	check(!Bunch.IsError());
	Ch->SendBunch( &Bunch, 0 );
	Bunch.Exit();
}

void UXC_ChannelDownload::Destroy()
{
	if( Ch && Ch->Download == (UChannelDownload*)this )
		Ch->Download = NULL;
	Ch = NULL;
	Super::Destroy();
}
IMPLEMENT_CLASS(UXC_ChannelDownload)



void UXC_FileChannel::Init( UNetConnection* InConnection, INT InChannelIndex, INT InOpenedLocally )
{
	Super::Init( InConnection, InChannelIndex, InOpenedLocally );
	if ( InConnection && InConnection->Driver && !InConnection->Driver->ServerConnection ) //This is not a client
		ChType = CHTYPE_File; //Avoid ULevel->NotifyAcceptingChannel from deleting this channel
}

void UXC_FileChannel::ReceivedBunch( FInBunch& Bunch )
{
//	UNREAL ADV: BUG REPORTED BY LUIGI AURIEMMA
//	check(!Closing);
	guard( UXC_FileChannel:ReceivedBunch);
	if ( Closing )
		return;
	if( OpenedLocally )
	{
		// Receiving a file sent from the other side.  If Bunch.GetNumBytes()==0, it means the server refused to send the file.
		Download->ReceiveData( Bunch.GetData(), Bunch.GetNumBytes() );
	}
	else
	{
		if( !Connection->Driver->AllowDownloads )
		{
			// Refuse the download by sending a 0 bunch.
			debugf( NAME_DevNet, LocalizeError(TEXT("NetInvalid"),TEXT("Engine")) );
			FOutBunch_Hack Bunch( this, 1 );
			SendBunch( &Bunch, 0 );
			Bunch.Exit();
			return;
		}
		if( SendFileAr )
		{
			FString Cmd;
			Bunch << Cmd;
			if( !Bunch.IsError() && Cmd==TEXT("SKIP") )
			{
				// User cancelled optional file download.
				// Remove it from the package map
				debugf( TEXT("User skipped download of '%s'"), SrcFilename );
				Connection->PackageMap->List.Remove( PackageIndex );
				return;
			}
		}
		else
		{
			// Request to send a file.
			FGuid Guid;
			Bunch << Guid;
			if( !Bunch.IsError() )
			{
				for( INT i=0; i<Connection->PackageMap->List.Num(); i++ )
				{
					FPackageInfo& Info = Connection->PackageMap->List(i);
					if( Info.Guid==Guid && Info.URL!=TEXT("") )
					{
						FixFilename( *Info.URL );
						if( Connection->Driver->MaxDownloadSize>0 && GXFileManager->FileSize(*Info.URL) > Connection->Driver->MaxDownloadSize )
							break;							
						appStrncpy( SrcFilename, *Info.URL, ARRAY_COUNT(SrcFilename) );
						if( Connection->Driver->Notify->NotifySendingFile( Connection, Guid ) )
						{
							check(Info.Linker);
							SendFileAr = NULL;
							FString FileToSend( SrcFilename);
							FileToSend += TEXT(".lzma");
							SendFileAr = GXFileManager->CreateFileReader( *FileToSend);
							if ( !SendFileAr )
							{
								FileToSend = SrcFilename;
								FileToSend += TEXT(".uz");
								SendFileAr = GXFileManager->CreateFileReader( *FileToSend);
							}
							if ( !SendFileAr )
								SendFileAr = GXFileManager->CreateFileReader( SrcFilename );
							if( SendFileAr )
							{
								// Accepted! Now initiate file sending.
								debugf( NAME_DevNet, LocalizeProgress(TEXT("NetSend"),TEXT("Engine")), *FileToSend );
								PackageIndex = i;
								return;
							}
						}
					}
				}
			}
		}

		// Illegal request; refuse it by closing the channel.
		debugf( NAME_DevNet, LocalizeError(TEXT("NetInvalid"),TEXT("Engine")) );
		
		FOutBunch_Hack Bunch( this, 1 );
		SendBunch( &Bunch, 0 );
		Bunch.Exit();
	}
	unguard;
}

void UXC_FileChannel::Tick()
{
	UChannel::Tick();
	Connection->TimeSensitive = 1;
	INT Size;

	//TIM: IsNetReady(1) causes the client's bandwidth to be saturated. Good for clients, very bad
	// for bandwidth-limited servers. IsNetReady(0) caps the clients bandwidth.
	static UBOOL LanPlay = ParseParam(appCmdLine(),TEXT("lanplay"));
	while( !OpenedLocally && SendFileAr && IsNetReady(LanPlay) && (Size=MaxSendBytes())!=0 )
	{
		// Sending.
		INT Remaining = SendFileAr->TotalSize() - SentData;
		Size = Min( Size, Remaining );
		//Never send less than 13 bytes, we ensure LZMA header is sent in one chunk
		if ( (SentData == 0) && (Size <= 13) )
			return;
		FOutBunch_Hack Bunch( this, Size>=Remaining );

		//Serialize directly INTO the bunch
//		SendFileAr->Serialize( Bunch.GetData(), Size);
		
		BYTE* Buffer = (BYTE*)appAlloca( Size );
		SendFileAr->Serialize( Buffer, Size );
		if( SendFileAr->IsError() )
		{
			//!!
		}
		SentData += Size;
		Bunch.Serialize( Buffer, Size );
		Bunch.Bunch()->bReliable = 1;
		check(!Bunch.IsError());
		SendBunch( &Bunch, 0 );
		Connection->FlushNet();
		if ( Bunch.Bunch()->bClose )
		{
			// Finished.
			delete SendFileAr;
			SendFileAr = NULL;
		}
		Bunch.Exit();
	}
}

void UXC_FileChannel::Destroy()
{
	check(Connection);
	if( RouteDestroy() )
		return;


	UChannel** Channels = !b440Net ? (UChannel**) &Connection->Channels : (UChannel**) ( (DWORD)Connection + 0x0E74 );
	check( Channels[ChIndex]==this);

	// Close the file.
	if( SendFileAr )
	{
		delete SendFileAr;
		SendFileAr = NULL;
	}

	// Notify that the receive succeeded or failed.
	if( OpenedLocally && Download )
	{
		Download->DownloadDone();
		delete Download;
	}
	UChannel::Destroy();
}
IMPLEMENT_CLASS(UXC_FileChannel)




//==============================================================================
// Net Connection Hack
//==============================================================================


//
// Setup any extra variables
//
void UXC_NetConnectionHack::StaticConstructor()
{
}


//
// Create a channel.
//
UChannel* UXC_NetConnectionHack::CreateChannel( EChannelType ChType, UBOOL bOpenedLocally, INT ChIndex )
{
	guard(UXC_NetConnectionHack::CreateChannel);
	check(UChannel::IsKnownChannelType(ChType));
	AssertValid();

	UChannel** Channels = GetChannels(); //Override global variable
	
	// If no channel index was specified, find the first available.
	if( ChIndex==INDEX_NONE )
	{
		INT FirstChannel = (ChType != CHTYPE_Control); //0 if control, 1 if other
		for( ChIndex=FirstChannel; ChIndex<MAX_CHANNELS; ChIndex++ )
			if( !Channels[ChIndex] )
				break;
		if( ChIndex==MAX_CHANNELS )
			return NULL;
	}

	// Make sure channel is valid.
	check(ChIndex<MAX_CHANNELS);
	check(Channels[ChIndex]==NULL);

	// Need to construct a non-unique name for this channel
	FName NewChannelName( *FString::Printf( TEXT("%s%i"), UChannel::ChannelClasses[ChType]->GetName(), ChIndex) );
	
	// Create channel. ** residing in the connection with the non-unique name
	UChannel* Channel = ConstructObject<UChannel>( UChannel::ChannelClasses[ChType], this, NewChannelName );
	Channel->Init( this, ChIndex, bOpenedLocally );
	Channels[ChIndex] = Channel;
	GetOpenChannels()->AddItem(Channel);

	return Channel;
	unguard;
}
//
// There are memory leaks with the data bunches, but hopefully minimal
//
void UXC_NetConnectionHack::Destroy()
{
	guard( XC_Core::~UXC_NetConnectionHack );
	Super::Destroy();
	//Fix for v451-v440
	if ( b440Net )
	{
		UNetConnection* Conn451 = (UNetConnection*) (((DWORD)this) + 20);
		if ( Conn451->QueuedAcks.GetData() )
			Conn451->QueuedAcks.Empty();
		if ( Conn451->ResendAcks.GetData() )
			Conn451->ResendAcks.Empty();
		if ( Conn451->OpenChannels.GetData() ) //Garbage collector will crash the game if we don't kill the channels
			Conn451->OpenChannels.Empty();
		if ( Conn451->SentTemporaries.GetData() )
			Conn451->SentTemporaries.Empty();
		if ( Conn451->ActorChannels.Num() > 0 )
			Conn451->ActorChannels.Empty();
		if ( Conn451->DownloadInfo.GetData() )
			Conn451->DownloadInfo.Empty();

#if DO_ENABLE_NET_TEST
		if ( Conn451->Delayed.GetData() )
			Conn451->Delayed.Empty();
#endif
		appMemzero( &Conn451->LastOut, 84); //FBitWriter size in v440
		appMemzero( &Conn451->Out, 84); //FBitWriter size in v440
		appMemzero( &Conn451->QueuedAcks, 132); //Clear dynamic arrays plus neg offsets
	}
	unguard;
}


IMPLEMENT_CLASS(UXC_NetConnectionHack)
